@@ -42,8 +42,10 @@
   :type 'string
   :group 'brltty)
 
-(defcustom brltty-default-port 35751
-  "Default BrlTTY port to connect to."
+(defcustom brltty-default-port '(4101 4102 4103 4104 35751)
+  "Default BrlTTY port to connect to.
+If it is a list, the given port numbers are attempted in the order they are
+given until Emacs connects to something."
   :type 'integer
   :group 'brltty)
 
@@ -85,7 +87,7 @@ available, from the  environment variable CONTROLVT."
 
 (defconst brltty--emacs-process-ok (fboundp 'process-put))
 
-(defconst brltty--protocol-version 7)
+(defconst brltty--supported-protocol-versions '(7))
 
 (defconst brltty--packet-types
   '(;; commands
@@ -109,6 +111,7 @@ available, from the  environment variable CONTROLVT."
 
 (defstruct brltty--connection
   process
+  (protocol-version nil)
   (display-width nil)
   (output "")
   (key-handler nil)
@@ -118,12 +121,23 @@ available, from the  environment variable CONTROLVT."
 (unless brltty--emacs-process-ok
   (defvar brltty--process-connections '()))
 
-(defun brltty--terminal-spec ()
+(lexical-let ((last-selected-frame 'uninitialized)
+              (last-terminal-spec 'uninitialized))
+  (defun brltty--terminal-spec ()
+    (if (eq (selected-frame) last-selected-frame)
+        last-terminal-spec
+      (prog1 (setq last-terminal-spec (brltty--terminal-spec*))
+        (setq last-selected-frame (selected-frame))))))
+
+(defun brltty--terminal-spec* ()
   (let ((terminal-spec '()))
     (cond
-     ((getenv "WINDOWSPATH")
+     ((or (getenv "WINDOWPATH")
+          (getenv "WINDOWSPATH"))
       (save-match-data
-        (dolist (number (split-string (getenv "WINDOWSPATH") ":"))
+        (dolist (number (split-string (or (getenv "WINDOWPATH")
+                                          (getenv "WINDOWSPATH"))
+                                      ":"))
           (push (string-to-number number) terminal-spec))))
      ((eq window-system 'x)
       (push brltty-tty terminal-spec)))
@@ -136,13 +150,14 @@ available, from the  environment variable CONTROLVT."
       (with-temp-buffer
         (let ((standard-input (current-buffer)))
           (push (or (ignore-errors
-                      (insert-file-contents "/proc/self/stat")
-                      (goto-char (point-min))
-                      (dotimes (_ 6) (read))
-                      (let ((tty (read)))
-                        (if (and tty (= (/ tty 256) 4)) ; major
-                            (mod tty 256)               ; minor
-                          0)))
+                      (with-speechd-coding-protection
+                        (insert-file-contents "/proc/self/stat")
+                        (goto-char (point-min))
+                        (dotimes (_ 6) (read))
+                        (let ((tty (read)))
+                          (if (and tty (= (/ tty 256) 4)) ; major
+                              (mod tty 256)               ; minor
+                            0))))
                     0)
                 terminal-spec))))
      (t
@@ -150,9 +165,19 @@ available, from the  environment variable CONTROLVT."
     (nreverse terminal-spec)))
 
 (defun brltty--open-connection (host port key-handler)
-  (let ((process (open-network-stream "brltty" nil
-                                      (or host brltty-default-host)
-                                      (or port brltty-default-port))))
+  (let ((process nil)
+        (ports (or port brltty-default-port))
+        (host* (or host brltty-default-host)))
+    (unless (consp ports)
+      (setq ports (list port)))
+    (while (and (not process)
+                ports)
+      (condition-case err
+          (setq process (open-network-stream "brltty" nil host* (car ports)))
+        (error
+         (if (cdr ports)
+             (setq ports (cdr ports))
+           (signal (car err) (cdr err))))))
     (when process
       (set-process-coding-system process 'binary 'binary)
       (if (fboundp 'set-process-query-on-exit-flag)
@@ -201,6 +226,12 @@ available, from the  environment variable CONTROLVT."
   (+ (* 256 256 256 (aref string 0)) (* 256 256 (aref string 1))
      (* 256 (aref string 2)) (aref string 3)))
 
+(defun brltty--read-integer64 (string)
+  (list
+   (+ (* 256 (aref string 0)) (aref string 1))
+   (+ (* 256 256 (aref string 2)) (* 256 (aref string 3)) (aref string 4))
+   (+ (* 256 256 (aref string 5)) (* 256 (aref string 6)) (aref string 7))))
+
 (defun brltty--read-packet (connection)
   (condition-case err
       (let ((process (brltty--connection-process connection)))
@@ -233,7 +264,11 @@ available, from the  environment variable CONTROLVT."
      (key
       (let ((handler (brltty--connection-key-handler connection)))
         (when handler
-          (funcall handler (brltty--read-integer data)))))
+          (funcall handler
+                   (funcall (if (> (length data) 4)
+                                #'brltty--read-integer64
+                              #'brltty--read-integer)
+                            data)))))
      (nil
       ;; unknown packet type -- ignore
       )
@@ -271,12 +306,13 @@ available, from the  environment variable CONTROLVT."
             (flet ((send-integer (n)
                      (process-send-string
                       process
-                      (apply #'format "%c%c%c%c"
-                             (funcall #'reverse
-                                      (loop for i from 1 to 4
-                                            for x = n then (/ x 256)
-                                            for rem = (% x 256)
-                                            collect rem))))))
+                      (string-make-unibyte
+                       (apply #'format "%c%c%c%c"
+                              (funcall #'reverse
+                                       (loop for i from 1 to 4
+                                             for x = n then (/ x 256)
+                                             for rem = (% x 256)
+                                             collect rem)))))))
               (send-integer length)
               (send-integer (cdr (assoc packet-id brltty--packet-types)))
               (dolist (data data-list)
@@ -285,7 +321,9 @@ available, from the  environment variable CONTROLVT."
                   (send-integer data))
                  ((vectorp data)
                   (dotimes (i (length data))
-                    (process-send-string process (format "%c" (aref data i)))))
+                    (process-send-string
+                     process
+                     (string-make-unibyte (format "%c" (aref data i))))))
                  (t
                   (process-send-string process data))))
               (when answer
@@ -313,20 +351,35 @@ available, from the  environment variable CONTROLVT."
 
 ;;; Public functions and data
 
+(put 'brltty-connection-error 'error-conditions
+     '(error speechd-connection-error brltty-connection-error))
+(put 'brltty-connection-error 'error-message
+     "Error on opening BrlTTY connection")
 
 (defun brltty-open (&optional host port key-handler)
   "Open and return connection to a BrlTTY server running on HOST and PORT.
 If HOST or PORT is nil, `brltty-default-host' or `brltty-default-port' is used
 respectively."
   (condition-case err
-      (let ((connection (brltty--open-connection host port key-handler)))
-        (brltty--send-packet connection 'ack 'authkey
-                             brltty--protocol-version
-                             (brltty--authentication-key))
+      (let ((connection nil)
+            (protocol-versions brltty--supported-protocol-versions))
+        (while (and (not connection)
+                    protocol-versions)
+          (let ((connection* (brltty--open-connection host port key-handler)))
+            (condition-case err*
+                (let ((version (car protocol-versions)))
+                  (brltty--send-packet connection* 'ack 'authkey version
+                                       (brltty--authentication-key))
+                  (setf (brltty--connection-protocol-version connection*)
+                        version)
+                  (setq connection connection*))
+              (error
+               (setq protocol-versions (cdr protocol-versions))
+               (unless protocol-versions
+                 (signal (car err*) (cdr err*)))))))
         connection)
     (error
-     (message "Error on opening BrlTTY connection: %s" err)
-     nil)))
+     (signal 'brltty-connection-error err))))
 
 (defun brltty-close (connection)
   "Close BrlTTY CONNECTION."
