@@ -34,6 +34,9 @@
 (require 'speechd-common)
 (require 'speechd-out)
 
+(require 'speechd-brltty)
+(require 'speechd-ssip)
+
 
 ;;; User options
 
@@ -100,7 +103,7 @@ See also `speechd-speak-buffer-insertions'."
   :type '(repeat (string :tag "Buffer name"))
   :group 'speechd-speak)
 
-(defcustom speechd-speak-priority-insertions-in-buffers '("*shell*")
+(defcustom speechd-speak-priority-insertions-in-buffers '()
   "List of names of buffers, in which insertions are spoken immediately.
 Unlike `speechd-speak-insertions-in-buffers', speaking is not delayed until a
 command is completed.
@@ -147,7 +150,9 @@ and the keys are read after the command is performed."
 
 (defcustom speechd-speak-ignore-command-keys
   '(forward-char backward-char next-line previous-line
-    delete-char delete-backward-char backward-delete-char-untabify)
+    delete-char comint-delchar-or-maybe-eof delete-backward-char
+    backward-delete-char-untabify
+    c-electric-backspace c-electric-delete-forward)
   "List of commands for which their keys are never read."
   :type '(repeat function)
   :group 'speechd-speak)
@@ -324,14 +329,18 @@ This may be useful when debugging speechd-el itself."
 
 (defvar speechd-speak--debug ())
 (defvar speechd-speak--max-debug-length 12)
+(defvar speechd-speak--debug-all nil)
+(defvar speechd-speak--debug-mark nil)
 
-(defun speechd-speak--debug (info)
-  (setq speechd-speak--debug
-        (cons info
-              (if (>= (length speechd-speak--debug)
-                      speechd-speak--max-debug-length)
-                  (butlast speechd-speak--debug)
-                speechd-speak--debug))))
+(defun speechd-speak--debug (info &optional optional)
+  (when (or (not optional) speechd-speak--debug-all)
+    (setq speechd-speak--debug
+          (cons info
+                (if (and (not speechd-speak--debug-all)
+                         (>= (length speechd-speak--debug)
+                             speechd-speak--max-debug-length))
+                    (butlast speechd-speak--debug)
+                  speechd-speak--debug)))))
 
 
 ;;; Control functions
@@ -464,11 +473,35 @@ This variable is reset to nil before each command in pre-command-hook.")
                                           speechd-default-text-priority)))
      ,@body))
 
+(defun speechd-speak--remove-invisible-text (text)
+  (let ((pieces '())
+        (max (length text))
+        (pos 0)
+        (invisibility-spec (if (eq buffer-invisibility-spec t)
+                               t
+                             (mapcar (lambda (s) (if (consp s) (car s) s))
+                                     buffer-invisibility-spec))))
+    (while (and pos (< pos max))
+      (let ((next-pos (next-single-char-property-change pos 'invisible text)))
+        (when (let ((invisibility (get-char-property pos 'invisible text)))
+                (not (cond
+		      ((eq invisibility-spec t)
+		       invisibility)
+		      ((consp invisibility)
+		       (intersection invisibility-spec invisibility))
+		      (t
+		       (memq invisibility buffer-invisibility-spec)))))
+          (push (substring text pos next-pos) pieces))
+        (setq pos next-pos)))
+    (apply 'concat (nreverse pieces))))
+
 (defun speechd-speak--text (text &rest args)
   (speechd-speak--maybe-speak
-   ;; TODO: skip invisible text
    ;; TODO: replace repeating patterns
    ;; TODO: handle selective display
+   (setq text (speechd-speak--remove-invisible-text text))
+   (when speechd-speak--debug-mark
+     (speechd-speak--debug (cons speechd-speak--debug-mark text) t))
    (apply #'speechd-out-text text args)))
 
 (defun speechd-speak--char (char &rest args)
@@ -587,11 +620,11 @@ If BUFFER is nil, read current buffer."
 
 (defun speechd-speak-read-mode-line ()
   "Read mode line.
-This function works only in Emacs 21.4 or higher."
+This function works only in Emacs 22 or higher."
   (interactive)
   (when (fboundp 'format-mode-line)
     (speechd-speak--interactive
-     (speechd-speak--text (format-mode-line t)))))
+     (speechd-speak--text (format-mode-line mode-line-format t)))))
 
 (defun speechd-speak--window-contents ()
   (sit-for 0)                           ; to update window start and end
@@ -638,6 +671,7 @@ This function works only in Emacs 21.4 or higher."
   modified
   (changes '())
   (change-end nil)
+  (deleted-chars nil)
   (other-changes '())
   other-changes-buffer
   other-window
@@ -712,12 +746,16 @@ This function works only in Emacs 21.4 or higher."
 
 (defmacro speechd-speak--defadvice (function class &rest body)
   (let* ((function* function)
-         (fname (if (listp function*) (first function*) function*))
-         (aname (if (listp function*) 'speechd-speak-user 'speechd-speak)))
-    `(defadvice ,fname (,class ,aname activate preactivate compile)
-       (if (not speechd-speak--started)
-           ,(when (eq class 'around) 'ad-do-it)
-         ,@body))))
+         (functions (if (listp function*) function* (list function*)))
+         (aname (if (listp function*) 'speechd-speak 'speechd-speak-user)))
+    `(progn
+       ,@(mapcar
+          (lambda (fname)
+            `(defadvice ,fname (,class ,aname activate preactivate compile)
+               (if (not speechd-speak--started)
+                   ,(when (eq class 'around) 'ad-do-it)
+                 ,@body)))
+          functions))))
 
 (defmacro speechd-speak--report (feedback &rest args)
   (if (stringp feedback)
@@ -786,6 +824,11 @@ Language must be an RFC 1766 language code, as a string."
   (speechd-set-language language)
   (setq speechd-language language))
 
+(defun speechd-speak--in-comint-p ()
+  "Return non-nil if the current buffer is any sort of a comint buffer."
+  (and (boundp 'comint-accum-marker)
+       comint-accum-marker))
+
 
 ;;; Basic speaking
 
@@ -826,23 +869,37 @@ Language must be an RFC 1766 language code, as a string."
 				 after
   (speechd-speak-read-word))
 
-(speechd-speak--command-feedback (delete-backward-char backward-delete-char
-				  backward-delete-char-untabify)
-				 around
-  (let ((deleted-char (preceding-char)))
-    (prog1 ad-do-it
-      (speechd-speak--with-updated-text
-        (speechd-speak-read-char (if speechd-speak-deleted-char
-                                     deleted-char
-                                   (preceding-char)))))))
+(defun speechd-speak--store-deleted-chars (text)
+  (speechd-speak--with-command-start-info
+    (setf (speechd-speak--cinfo deleted-chars)
+          (if (speechd-speak--cinfo deleted-chars)
+              t
+            text))))
 
-(speechd-speak--command-feedback (delete-char) around
-  (let ((deleted-char (following-char)))
-    (prog1 ad-do-it
-      (speechd-speak--with-updated-text
-       (speechd-speak-read-char (if speechd-speak-deleted-char
-                                    deleted-char
-                                  (following-char)))))))
+(speechd-speak--defadvice (delete-backward-char backward-delete-char) around
+  (speechd-speak--with-command-start-info
+    (let ((speechd-speak--deleted-chars
+           (when speechd-speak-deleted-char
+             (speechd-speak--buffer-substring
+              (point)
+              (max (- (point) (or (ad-get-arg 0) 1))
+                   (point-min))))))
+      ad-do-it
+      (speechd-speak--store-deleted-chars (if speechd-speak-deleted-char
+                                              speechd-speak--deleted-chars
+                                            (format "%c" (preceding-char)))))))
+  
+(speechd-speak--defadvice (delete-char) around
+  (let ((speechd-speak--deleted-chars
+         (when speechd-speak-deleted-char
+           (speechd-speak--buffer-substring
+            (point)
+            (min (+ (point) (or (ad-get-arg 0) 1))
+                 (point-max))))))
+    ad-do-it
+    (speechd-speak--store-deleted-chars (if speechd-speak-deleted-char
+                                            speechd-speak--deleted-chars
+                                          (format "%c" (following-char))))))
 
 (speechd-speak--command-feedback (quoted-insert) after
   (speechd-speak-read-char (preceding-char)))
@@ -853,15 +910,6 @@ Language must be an RFC 1766 language code, as a string."
 (speechd-speak--command-feedback (undo) after
   (speechd-speak-read-line))
 
-(defmacro speechd-speak--unhide-message (function)
-  ;; If `message' is invoked within a built-in function, there's no way to get
-  ;; notified automatically about it.  So we have to wrap the built-in
-  ;; functions displaying messages to check for the otherwise hidden messages.
-  `(speechd-speak--defadvice ,function after
-     (speechd-speak--current-message)))
-
-(speechd-speak--unhide-message write-region)
-
 
 ;;; Killing and yanking
 
@@ -948,6 +996,16 @@ Language must be an RFC 1766 language code, as a string."
 (speechd-speak--defadvice message after
   (speechd-speak--current-message))
 
+(defmacro speechd-speak--unhide-message (function)
+  ;; If `message' is invoked within a built-in function, there's no way to get
+  ;; notified automatically about it.  So we have to wrap the built-in
+  ;; functions displaying messages to check for the otherwise hidden messages.
+  `(speechd-speak--defadvice ,function after
+     (unless (string= (current-message) speechd-speak--last-message)
+       (speechd-speak--current-message))))
+
+(speechd-speak--unhide-message write-region)
+
 
 ;;; Minibuffer
 
@@ -964,6 +1022,10 @@ Language must be an RFC 1766 language code, as a string."
                        :icon (unless no-icon 'minibuffer)))
 
 (defun speechd-speak--speak-minibuffer-prompt ()
+  (speechd-speak--with-command-start-info
+   ;; Discard changes, otherwise they would be read *after* the prompt, perhaps
+   ;; confusing the user
+   (setf (speechd-speak--cinfo changes) '()))
   (let ((speechd-language "en")
         (speechd-speak-input-method-languages nil))
     (speechd-speak--prompt (minibuffer-prompt)))
@@ -1142,11 +1204,13 @@ Only single characters are allowed in the keymap.")
                  beg)
                end)))
     (dolist (o (overlays-in beg end))
-      (let ((face (overlay-get o 'face)))
-        (when face
+      (let ((face (overlay-get o 'face))
+            (invisible (overlay-get o 'invisible)))
+        (when (or face invisible)
           (let ((beg* (max (overlay-start o) beg))
                 (end* (min (overlay-end o) end)))
-            (put-text-property (- beg* beg) (- end* beg) 'face face text)))))
+            (add-text-properties (- beg* beg) (- end* beg)
+                                 `(face ,face invisible ,invisible) text)))))
     text))
 
 (defun speechd-speak--minibuffer-update-report (info old new)
@@ -1175,19 +1239,32 @@ Only single characters are allowed in the keymap.")
                            (current-buffer))
     (speechd-speak--text text :priority 'message)))
 
+(defvar speechd-speak--current-change-string nil)
+(defun speechd-speak--before-change-hook (beg end)
+  ;; Beware, before-change-functions may be sometimes called for several times
+  ;; with different arguments (an Emacs bug?)
+  (unless speechd-speak--current-change-string
+    (setq speechd-speak--current-change-string
+          (buffer-substring-no-properties beg end))))
 (defun speechd-speak--after-change-hook (beg end len)
   (speechd-speak--enforce-speak-mode)
   (speechd-speak--with-command-start-info
-    (unless (= beg end)
+    (unless (or (= beg end)
+                ;; Avoid reading changes when only text properties have changed
+                (equal speechd-speak--current-change-string
+                       (buffer-substring-no-properties beg end)))
       (cond
        ((or (member (buffer-name) speechd-speak-priority-insertions-in-buffers)
+            (speechd-speak--in-comint-p)
             ;; Asynchronous buffer changes
             (and (not this-command)
                  (member (buffer-name) speechd-speak-insertions-in-buffers)))
         (unless (memq this-command '(self-insert-command quoted-insert newline
                                      newline-and-indent undo yank yank-pop))
+          ;; Don't align buffer substring here so that the last word of comint
+          ;; input is not read
           (speechd-speak--read-buffer-change
-           (buffer-name) (speechd-speak--buffer-substring beg end t))))
+           (buffer-name) (speechd-speak--buffer-substring beg end nil))))
        ((not this-command)
         ;; Asynchronous buffer change -- we are not interested in it by
         ;; default
@@ -1201,7 +1278,8 @@ Only single characters are allowed in the keymap.")
             (progn
               (speechd-speak--read-other-changes)
               (speechd-speak--minibuffer-update beg end len))
-          (speechd-speak--add-command-text info beg end)))))))
+          (speechd-speak--add-command-text info beg end))))))
+  (setq speechd-speak--current-change-string nil))
 
 (defconst speechd-speak--dont-cancel-on-commands
   '(speechd-speak speechd-unspeak
@@ -1234,7 +1312,7 @@ Only single characters are allowed in the keymap.")
                 (speechd-speak--prompt (match-string 2 interactive)))))))
     (error
      (speechd-speak--debug (list 'pre-command-hook-error err))
-     (apply #'error (cdr err))))
+     (signal (car err) (cdr err))))
   (add-hook 'pre-command-hook 'speechd-speak--pre-command-hook))
 
 (defmacro speechd-speak--post-defun (name shy new-state guard &rest body)
@@ -1247,7 +1325,7 @@ Only single characters are allowed in the keymap.")
                    in-minibuffer other-buffer)
        (if (and ,state-condition
                 ,guard)
-           (progn
+           (let ((speechd-speak--debug-mark (quote ,name)))
              ,@body
              ,new-state)
          state))))
@@ -1274,6 +1352,23 @@ Only single characters are allowed in the keymap.")
   (when (memq speechd-speak-buffer-name '(text nil))
     (speechd-speak-read-line t)))
 
+(speechd-speak--post-defun deleted-chars t t
+  ;; Reading delete-char etc. interactive feedback
+  (let ((deleted-chars (speechd-speak--cinfo deleted-chars))
+        (changes (speechd-speak--cinfo changes)))
+    (and (stringp deleted-chars)
+         (or (null changes)
+             (and (= (length changes) 1)
+                  (equal (first changes) deleted-chars)))))
+  (when (eq speechd-speak-read-command-keys t)
+    ;; Cancel reading the DEL key etc. -- perhaps too daring?
+    (speechd-out-cancel))
+  (speechd-speak--with-updated-text
+   (let ((deleted-chars (speechd-speak--cinfo deleted-chars)))
+     (if (= (length deleted-chars) 1)
+         (speechd-speak-read-char (string-to-char deleted-chars))
+       (speechd-speak--text deleted-chars)))))
+
 (speechd-speak--post-defun command-keys t nil
   ;; Keys that invoked the command
   (and (not (memq this-command speechd-speak-ignore-command-keys))
@@ -1452,6 +1547,7 @@ Only single characters are allowed in the keymap.")
 (defvar speechd-speak--post-command-speaking-defaults
   '(speechd-speak--post-read-special-commands
     speechd-speak--post-read-buffer-switch
+    speechd-speak--post-read-deleted-chars
     speechd-speak--post-read-command-keys
     speechd-speak--post-read-command-done
     speechd-speak--post-read-info-change
@@ -1514,7 +1610,7 @@ Only single characters are allowed in the keymap.")
              (setq speechd-speak--last-minibuffer-depth (minibuffer-depth)))))
         (error
          (speechd-speak--debug (list 'post-command-hook-top-error err))
-         (apply #'error (cdr err))))
+         (signal (car err) (cdr err))))
     (add-hook 'post-command-hook 'speechd-speak--post-command-hook)))
 
 
@@ -1551,7 +1647,8 @@ Only single characters are allowed in the keymap.")
   (speechd-speak--speak-completion))
 
 (speechd-speak--defadvice widget-choose around
-  (let ((widget-menu-minibuffer-flag t))
+  (let ((widget-menu-minibuffer-flag (or speechd-speak-mode
+                                         widget-menu-minibuffer-flag)))
     ad-do-it))
 
 
@@ -1713,7 +1810,7 @@ When the mode is enabled, all spoken text is spelled."
 (speechd-speak--watch header-line
   #'(lambda ()
       (if (fboundp 'format-mode-line)
-          (let ((line (format-mode-line t)))
+          (let ((line (format-mode-line header-line-format)))
             (if (string= line "") "empty" line))
         "unknown"))
   :on-change #'(lambda (old new)
@@ -1919,6 +2016,7 @@ the value of the `speechd-speak-prefix' variable:
         (speechd-speak-map-mode 1)
         (add-hook 'pre-command-hook 'speechd-speak--pre-command-hook)
         (add-hook 'post-command-hook 'speechd-speak--post-command-hook)
+        (add-hook 'before-change-functions 'speechd-speak--before-change-hook)
         (add-hook 'after-change-functions 'speechd-speak--after-change-hook)
         (add-hook 'minibuffer-setup-hook 'speechd-speak--minibuffer-setup-hook)
         (add-hook 'minibuffer-exit-hook 'speechd-speak--minibuffer-exit-hook)
@@ -1978,6 +2076,7 @@ starts blocking your Emacs functions."
     (setq speechd-speak--message-timer nil))
   (remove-hook 'pre-command-hook 'speechd-speak--pre-command-hook)
   (remove-hook 'post-command-hook 'speechd-speak--post-command-hook)
+  (remove-hook 'before-change-functions 'speechd-speak--before-change-hook)
   (remove-hook 'after-change-functions 'speechd-speak--after-change-hook)
   (remove-hook 'minibuffer-setup-hook 'speechd-speak--minibuffer-setup-hook)
   (remove-hook 'minibuffer-exit-hook 'speechd-speak--minibuffer-exit-hook)
@@ -1998,7 +2097,11 @@ starts blocking your Emacs functions."
   (global-speechd-speak-mode 1)
   (global-speechd-speak-map-mode 1)
   (speechd-speak--debug 'start)
-  (speechd-speak-report 'start)
+  (condition-case err (speechd-speak-report 'start)
+    (speechd-connection-error
+     (message (format "%s: %s" (car err) (cdr err)))
+     (sit-for 2)
+     (message "Customize `speechd-out-active-drivers' to disable drivers you don't use.")))
   (setq speechd-speak--message-timer
         (run-with-idle-timer 0 t 'speechd-speak--message-timer))
   (run-hooks 'speechd-speak-hook))
