@@ -198,6 +198,11 @@ You must reopen the connections to apply the changes to this variable."
              (speechd-reopen)))           
   :group 'speechd)
 
+(defcustom speechd-cancelable-connections '()
+  "List of names of connections on which cancel operation applies by default."
+  :type '(repeat (string :tag "Connection name"))
+  :group 'speechd)
+
 (defcustom speechd-face-voices '()
   "Alist mapping faces to voices.
 Each of the alist element is of the form (FACE . VOICE) where FACE is a face
@@ -330,7 +335,8 @@ current voice."
 
 (defmacro speechd--with-current-connection (&rest body)
   `(let ((connection (speechd--connection)))
-     ,@body))
+     (unless (speechd--connection-failure-p connection)
+       ,@body)))
 
 (defmacro speechd--with-connection-setting (var value &rest body)
   (let ((accessor (intern (concat "speechd--connection-" (symbol-name var))))
@@ -481,13 +487,20 @@ current voice."
                              speechd--end-regexp
                              (speechd--connection-process-output connection)))
                   (unless (accept-process-output process speechd-timeout nil 1)
-                    (error "Timeout in communication with Speech Dispatcher"))))
+                    (signal 'ssip-connection-error
+                            "Timeout in communication with Speech Dispatcher"))))
               (prog1 (speechd--connection-process-output connection)
                 (setf (speechd--connection-process-output connection) "")))
           (error
            (speechd--permanent-connection-failure connection)
-           (error "Error in communication with Speech Dispatcher")))))))
-  
+           (signal 'ssip-connection-error
+                   "Error in communication with Speech Dispatcher")))))))
+
+(put 'ssip-connection-error 'error-conditions
+     '(error speechd-connection-error ssip-connection-error))
+(put 'ssip-connection-error 'error-message
+     "Error on opening Speech Dispatcher connection")
+
 ;;;###autoload
 (defun* speechd-open (&optional host port &key quiet force-reopen)
   "Open connection to Speech Dispatcher running on the given host and port.
@@ -527,19 +540,25 @@ Return the opened connection on success, nil otherwise."
                             default-parameters))
                           (t
                            default-parameters)))
-	     (id (when (or (not connection)
-                           (not (speechd--connection-failure-p connection))
-                           force-reopen)
-                   (condition-case nil
-                       (speechd--open-connection host port)))))
-	(if id
-	    (progn
-              (setq connection (make-speechd--connection
-                                :name name :host host :port port :process id))
-              (puthash name connection speechd--connections))
-	  (unless quiet
-            (setq connection nil)
-	    (message "Connection to Speech Dispatcher failed")))
+             (connection-error nil)
+	     (id (cond
+		  ((or (not connection)
+		       (not (speechd--connection-failure-p connection))
+		       force-reopen)
+                   (condition-case err (speechd--open-connection host port)
+                     (file-error
+                      (setq connection-error err)
+                      nil)))
+		  ((and connection
+			(not (speechd--connection-failure-p connection)))
+		   (speechd--connection-id connection)))))
+        (when (or id (not quiet))
+          (setq connection (make-speechd--connection
+                            :name name :host host :port port :process id))
+          (puthash name connection speechd--connections))
+        (when (and (not id) (not quiet))
+          (speechd--permanent-connection-failure connection)
+          (signal 'ssip-connection-error connection-error))
 	(when id
 	  (speechd--set-connection-name name)
           (setq parameters (append parameters
@@ -989,6 +1008,7 @@ of the symbols `important', `message', `text', `notification' or
                       ((= character ? ) "space")
                       ((= character ?_) "underscore")
                       ((= character ?_) "double-quote")
+                      ((= character 127) "backspace")
                       (t (format "%c" character)))
                    (format "%s" character))))
     (dolist (m modifiers)
@@ -1011,8 +1031,15 @@ of the symbols `important', `message', `text', `notification' or
                                   (speechd--connection)))
                           command)))
       (speechd--send-command (list command "self"))))
+   ((eq all 'some)
+    (let ((speechd-client-name$ speechd-client-name))
+      (speechd--iterate-clients
+       (when (or (equal speechd-client-name$ speechd-client-name)
+                 (member speechd-client-name speechd-cancelable-connections))
+         (speechd--control-command command nil repeatable)))))
    ((numberp all)
-    (speechd--iterate-clients (speechd--control-command command nil)))
+    (speechd--iterate-clients 
+     (speechd--control-command command nil repeatable)))
    (t
     (speechd--send-command (list command "all")))))
 
@@ -1021,9 +1048,12 @@ of the symbols `important', `message', `text', `notification' or
   "Stop speaking all the messages sent through the current client so far.
 If the universal argument is given, stop speaking messages of all clients.
 If a numeric argument is given, stop speaking messages of all current Emacs
-session clients."
+session clients.
+If no argument is given, stop speaking messages of the current client and all
+the clients of the current Emacs session named in
+`speechd-cancelable-connections'."
   (interactive "P")
-  (speechd--control-command "CANCEL" all))
+  (speechd--control-command "CANCEL" (or all 'some) t))
 
 ;;;###autoload
 (defun speechd-stop (&optional all)
@@ -1041,7 +1071,8 @@ If the optional argument ALL is non-nil, pause speaking in all clients."
   (if all
       (speechd--iterate-connections
         (setf (speechd--connection-paused-p connection) t))
-    (setf (speechd--connection-paused-p (speechd--connection)) t))
+    (speechd--with-current-connection
+      (setf (speechd--connection-paused-p connection) t)))
   (speechd--control-command "PAUSE" (not (not all))))
 
 ;;;###autoload
@@ -1050,12 +1081,13 @@ If the optional argument ALL is non-nil, pause speaking in all clients."
 If the optional argument ALL is non-nil, resume speaking messages of all
 clients."
   (interactive "P")
-  (when (or all (speechd--connection-paused-p (speechd--connection)))
-    (speechd--control-command "RESUME" (not (not all)))
-    (if all
-        (setf (speechd--connection-paused-p (speechd--connection)) nil)
-      (speechd--iterate-connections
-        (setf (speechd--connection-paused-p connection) nil)))))
+  (speechd--with-current-connection
+    (when (or all (speechd--connection-paused-p connection))
+      (speechd--control-command "RESUME" (not (not all)))
+      (if all
+          (setf (speechd--connection-paused-p connection) nil)
+        (speechd--iterate-connections
+          (setf (speechd--connection-paused-p connection) nil))))))
 
 ;;;###autoload
 (defun speechd-repeat ()
